/*	
 *	clipptp.c
 *
 *	USB PTP Responder (Picture Transfer Protocol)
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2014 Naoyuki Sawa
 *
 *	* Fri Sep 05 21:36:08 JST 2014 Naoyuki Sawa
 *	- 1st [XB
 *	* Sat Sep 06 23:41:20 JST 2014 Naoyuki Sawa
 *	- Xg[ŃG[ʒḿAł悤ɂȂ܂B
 *	  ڍׂ́AuG[̕@ɂāv́u2014/09/06ǋLvQƂĂB
 */
#include "clip.h"

/****************************************************************************
 *	Ql
 ****************************************************************************/
//G[̕@ɂ
//uPTP Error Recoveryv(http://msdn.microsoft.com/en-us/library/windows/hardware/ff546442(v=vs.85).aspx)p:
//During initialization of the Microsoft PTP class minidriver (that is, on initial retrieval of the DeviceInfo and ObjectInfo datasets, and the property descriptions),
//any error is treated as a catastrophic failure, and the WIA minidriver fails to initialize.
//During later processing (for example, while retrieving an image), when an unrecognized error occurs,
//the Microsoft PTP minidriver first attempts to send the Get Device Status USB class-specific request (described in the USB Still Image Capture Device Definition).
//If that request succeeds, the driver clears any stalled endpoints and continues.
//If the Get Device Status request fails, the PTP minidriver attempts to reset the device using the Device Reset class-specific request (described in the USB Still Image Capture Device Definition).
//If the Device Reset class-specific request succeeds, it returns S_FALSE instead of S_OK. If resetting the device fails, the Device Reset class-specific request returns an error code.
//ƏĂ邪AۂɎƁAG[̕oȂB
//̓Iɂ́AGet Device StatusDevice Reset͌Ă΂邪ǍClear FeatureĂ΂ȂB
//̌́AGet Device Status̖߂l̍\AȂ̂ƎvB
//{AGet Device Status̖߂lɂ́AXg[ĂpCvǉĕԂƂɂȂĂ̂A
//dl(Universal Serial Bus Still Image Capture Device Definition)̐BŁA̓Iȍ\킩ȂB
//ŁA1oCgzŃoNGh|Cg̔ԍ(0x020x82)i[Ă݂AClear FeatureĂ΂ȂB
//̖ڏȀ݂AXg[ŃG[ʒḿAP/ECE𔲂邵̕@B
//ł邾Xg[ɁAX|XtF[YŃG[Ԃ悤ɍHv邵ƎvB
//2014/09/06ǋL
//܂B܂Ő삵ĂȂ̂́Aȉ̓̌łB
//1. clipusb.cdevice_request_endpoint()ɃoOLAclear_feature()삵Ă܂łB
//2. Get Device Status Request̖߂lParameterUINT8zɂĂ܂BUINT32złB
//LCƂAXg[ŃG[ʒmɁAĎ̏s悤ɂȂ܂B
/****************************************************************************
 *	O[oϐ
 ****************************************************************************/
ST_PTP stPTP;
/****************************************************************************
 *	[J֐錾
 ****************************************************************************/
static void get_descriptor(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
static void set_configuration(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
static void set_interface(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
static void class_request(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
static void ep2_rxdone();
static void ep2_txdone();
static int ep2_txsub();
static void put_string(void* _dst, const char* src);
static void get_string(char* dst, const void* _src);
static int is_protected_filename(const char* name);
static int process_command();
static int process_dataout(const void* buf, int len);
static int process_datain(void* buf, int len);
/****************************************************************************
 *	USBtbN
 ****************************************************************************/
static const USBINF usb_inf = {
	/* USBC mode */
	mode:			D12_ENDP_NONISO,
	/* USBC raw event */
	ep2_rxdone:		ep2_rxdone,
	ep2_txdone:		ep2_txdone,
	/* Device request */
	get_descriptor:		get_descriptor,
	set_configuration:	set_configuration,
	set_interface:		set_interface,
	/* Class request */
	class_request:		class_request,
};
/****************************************************************************
 *	fBXNv^
 ****************************************************************************/
/* D.6.1.1 Device Descriptor */
static const DEVICE_DESCRIPTOR device_desc = {
	bLength:		sizeof(DEVICE_DESCRIPTOR),		//The length of this descriptor
	bDescriptorType:	DEVICE_DESCRIPTOR_TYPE,			//Device Descriptor Type
	bcdUSB:			0x0110,					//USB Specification Release Number
//sv	bDeviceClass:		0x00,					//Class information may be found at the interface level
//sv	bDeviceSubClass:	0x00,					//bDeviceSubClass is zero
//sv	bDeviceProtocol:	0x00,					//bDeviceProtocol is zero
	bMaxPacketSize0:	EP0_PACKET_SIZE,			//mplementation specific, may be set to 8, 16, 32 or 64
	idVendor:		PCE_VENDOR_ID,				//Vendor ID assigned to IHV by USB-IF
	idProduct:		PTP_PRODUCT_ID,				//Product ID assigned by IHV
	bcdDevice:		0x0001,					//Device release number in BCD assigned by IHV to this release of model
	iManufacture:		1,					//Index of string descriptor describing IHV. If set to zero (0), there is no manufacturer string.
	iProduct:		2,					//Index of string describing this product. If set to zero (0), there is no product string.
//sv	iSerialNumber:		0,					//Index of string descriptor with this specific devicefs serial number. If set to zero (0), this device does not have a serial number.
	bNumConfigurations:	1,					//Device has this number of configurations.
};
/*--------------------------------------------------------------------------*/
/* D.6.1.2 Configuration Descriptor */
static const CONFIGURATION_DESCRIPTOR config_desc = {
	bLength:		sizeof(CONFIGURATION_DESCRIPTOR),	//The length of this descriptor
	bDescriptorType:	CONFIGURATION_DESCRIPTOR_TYPE,		//Configuration Descriptor Type
	wTotalLength:		sizeof(CONFIGURATION_DESCRIPTOR) +	//Total length of data returned for this configuration. Includes this descriptor and all of the descriptors that follow (interface, endpoint and vendor specific, if present).
				sizeof(INTERFACE_DESCRIPTOR) +
				sizeof(ENDPOINT_DESCRIPTOR) * 3,
	bNumInterfaces:		1,					//This configuration has <#> interfaces
	bConfigurationValue:	1,					//Implementation specific value used as an argument to Set Configuration to select this configuration
//sv	iConfiguration:		0,					//Index of string describing this configu ration. If set to zero (0), there is no configuration string.
	bmAttributes:		0x80,		/* Bus-powered */	//Configuration characteristics as defined by USB Specification
	MaxPower:		100 / 2,	/* 100[mA]() */	//Maximum power draw from the bus by this device when this configuration is selected.
};
/* D.6.1.3 Still Image Interface Descriptor */
static const INTERFACE_DESCRIPTOR interface_desc = {
	bLength:		sizeof(INTERFACE_DESCRIPTOR),		//The length of this descriptor
	bDescriptorType:	INTERFACE_DESCRIPTOR_TYPE,		//Interface Descriptor Type
//sv	bInterfaceNumber:	0,					//Number of interface in configuration
//sv	bAlternateSetting:	0x00,					//Default setting for this interface
	bNumEndpoints:		3,					//Number of endpoints in this interface, not including the default endpoint.
	bInterfaceClass:	0x06,					//Image interface
	bInterfaceSubClass:	0x01,					//Still Image Capture Device
	bInterfaceProtocol:	0x01,					//This field indicates the protocol sup ported by this device. PIMA 15740 compliant devices use Bulk-only protocol
//sv	iInterface:		0,					//Index of string describing this inter face. If set to zero (0), there is no inter face string.
};
/* D.6.1.4 Data-In Endpoint Descriptor */
static const ENDPOINT_DESCRIPTOR endpoint_desc_ep2in = {
	bLength:		sizeof(ENDPOINT_DESCRIPTOR),		//The length of this descriptor
	bDescriptorType:	ENDPOINT_DESCRIPTOR_TYPE,		//Endpoint Descriptor Type
	bEndpointAddress:	0x82,		/* EP2IN */		//Any endpoint number not used by another endpoint. The direction shall be set to IN.
	bAttributes:		2,		/* Bulk */		//The endpoint uses bulk transfers
	wMaxPacketSize:		EP2_NONISO_PACKET_SIZE,			//Maximum packet size used by this endpoint. This must be less than or equal to 64 bytes.
//sv	bInterval:		0,					//Ignored
};
/* D.6.1.5 Data-Out Endpoint Descriptor */
static const ENDPOINT_DESCRIPTOR endpoint_desc_ep2out = {
	bLength:		sizeof(ENDPOINT_DESCRIPTOR),		//The length of this descriptor
	bDescriptorType:	ENDPOINT_DESCRIPTOR_TYPE,		//Endpoint Descriptor Type
	bEndpointAddress:	0x02,		/* EP2OUT */		//Any endpoint number not used by another endpoint. The direction shall be set to OUT.
	bAttributes:		2,		/* Bulk */		//The endpoint uses bulk transfers
	wMaxPacketSize:		EP2_NONISO_PACKET_SIZE,			//Maximum packet size used by this endpoint. This must be less than or equal to 64 bytes.
//sv	bInterval:		0,					//Ignored
};
/* D.6.1.6 Interrupt Endpoint Descriptor */
static const ENDPOINT_DESCRIPTOR endpoint_desc_ep1in = {
	bLength:		sizeof(ENDPOINT_DESCRIPTOR),		//The length of this descriptor
	bDescriptorType:	ENDPOINT_DESCRIPTOR_TYPE,		//Endpoint Descriptor Type
	bEndpointAddress:	0x81,		/* EP1IN */		//Any endpoint number not used by another endpoint. The direction shall be set to IN.
	bAttributes:		3,		/* Interrupt */		//The endpoint uses interrupt transfers
	wMaxPacketSize:		EP1_PACKET_SIZE,			//Maximum packet size used by this endpoint. This must be less than or equal to 64 bytes.
	bInterval:		100,		/* 100[ms] */		//Interval for polling endpoint for data transfers
};
/*--------------------------------------------------------------------------*/
/* D.6.1.7 String Descriptors */
static const char* const string_desc[] = {
//	   ++-- 04 06 08 0a 0c 0e 10 12 14 16 18 1a 1c 1e 20 22 24 26 28 2a 2c 2e 30
//	   ||
//	"\x??\3_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0"
	"\x04\3\x09\x04",						// 0: wLANGID = 0x0409 English (United States)
	"\x16\3P\0i\0e\0c\0e\0 \0L\0a\0b\0.\0",				// 1: iManufacture   = "Piece Lab."
	"\x22\3P\0/\0E\0C\0E\0 \0U\0S\0B\0 \0d\0e\0v\0i\0c\0e\0",	// 2: iProduct       = "P/ECE USB device"
};
/****************************************************************************
 *	̑̃[Jf[^
 ****************************************************************************/
/* R}htF[Y̎ɍsׂtF[Y */
static const uint8_t TBL_DataDirection[] = {
	PTP_Phase_DataIn,						//0x1001 GetDeviceInfo
	PTP_Phase_Response,						//0x1002 OpenSession
	PTP_Phase_Response,						//0x1003 CloseSession
	PTP_Phase_DataIn,						//0x1004 GetStorageIDs
	PTP_Phase_DataIn,						//0x1005 GetStorageInfo
	PTP_Phase_Response,						//0x1006 GetNumObjects
	PTP_Phase_DataIn,						//0x1007 GetObjectHandles
	PTP_Phase_DataIn,						//0x1008 GetObjectInfo
	PTP_Phase_DataIn,						//0x1009 GetObject
	PTP_Phase_DataIn,						//0x100A GetThumb
	PTP_Phase_Response,						//0x100B DeleteObject
	PTP_Phase_DataOut,						//0x100C SendObjectInfo
	PTP_Phase_DataOut,						//0x100D SendObject
	PTP_Phase_Response,						//0x100E InitiateCapture
	PTP_Phase_Response,						//0x100F FormatStore
	PTP_Phase_Response,						//0x1010 ResetDevice
	PTP_Phase_Response,						//0x1011 SelfTest
	PTP_Phase_Response,						//0x1012 SetObjectProtection
	PTP_Phase_Response,						//0x1013 PowerDown
	PTP_Phase_DataIn,						//0x1014 GetDevicePropDesc
	PTP_Phase_DataIn,						//0x1015 GetDevicePropValue
	PTP_Phase_DataOut,						//0x1016 SetDevicePropValue
	PTP_Phase_Response,						//0x1017 ResetDevicePropValue
	PTP_Phase_Response,						//0x1018 TerminateOpenCapture
	PTP_Phase_Response,						//0x1019 MoveObject
	PTP_Phase_Response,						//0x101A CopyObject
	PTP_Phase_DataIn,						//0x101B GetPartialObject
	PTP_Phase_Response,						//0x101C InitiateOpenCapture
};
/*--------------------------------------------------------------------------*/
/* DeviceInfo Dataset */
static const uint8_t DeviceInfoDataset[] = {
	PTP_UINT16(100),						// 1 StandardVersion		2 UINT16	u0x0100vł͂Ȃu100vŐB
	PTP_UINT32(0),							// 2 VendorExtensionID		4 UINT32
	PTP_UINT16(0),							// 3 VendorExtensionVersion	2 UINT16
	0,								// 4 VendorExtensionDesc	? String
	PTP_UINT16(0),							// 5 FunctionalMode		2 UINT16
	PTP_UINT32(12),							// 6 OperationsSupported	? OperationCode Array
	 PTP_UINT16(PTP_OperationCode_GetDeviceInfo),		// 1
	 PTP_UINT16(PTP_OperationCode_OpenSession),		// 2
	 PTP_UINT16(PTP_OperationCode_CloseSession),		// 3
	 PTP_UINT16(PTP_OperationCode_GetStorageIDs),		// 4
	 PTP_UINT16(PTP_OperationCode_GetStorageInfo),		// 5
	 PTP_UINT16(PTP_OperationCode_GetNumObjects),		// 6
	 PTP_UINT16(PTP_OperationCode_GetObjectHandles),	// 7
	 PTP_UINT16(PTP_OperationCode_GetObjectInfo),		// 8
	 PTP_UINT16(PTP_OperationCode_GetObject),		// 9
//	 PTP_UINT16(PTP_OperationCode_GetThumb),		//
	 PTP_UINT16(PTP_OperationCode_DeleteObject),		//10
	 PTP_UINT16(PTP_OperationCode_SendObjectInfo),		//11
	 PTP_UINT16(PTP_OperationCode_SendObject),		//12
//	 PTP_UINT16(PTP_OperationCode_InitiateCapture),		//
//	 PTP_UINT16(PTP_OperationCode_FormatStore),		//
//	 PTP_UINT16(PTP_OperationCode_ResetDevice),		//
//	 PTP_UINT16(PTP_OperationCode_SelfTest),		//
//	 PTP_UINT16(PTP_OperationCode_SetObjectProtection),	//
//	 PTP_UINT16(PTP_OperationCode_PowerDown),		//
//	 PTP_UINT16(PTP_OperationCode_GetDevicePropDesc),	//
//	 PTP_UINT16(PTP_OperationCode_GetDevicePropValue),	//
//	 PTP_UINT16(PTP_OperationCode_SetDevicePropValue),	//
//	 PTP_UINT16(PTP_OperationCode_ResetDevicePropValue),	//
//	 PTP_UINT16(PTP_OperationCode_TerminateOpenCapture),	//
//	 PTP_UINT16(PTP_OperationCode_MoveObject),		//
//	 PTP_UINT16(PTP_OperationCode_CopyObject),		//
//	 PTP_UINT16(PTP_OperationCode_GetPartialObject),	//
//	 PTP_UINT16(PTP_OperationCode_InitiateOpenCapture),	//
	PTP_UINT32(0),							// 7 EventsSupported		? EventCode Array
	PTP_UINT32(0),							// 8 DevicePropertiesSupported	? DevicePropCode Array
	PTP_UINT32(0),							// 9 CaptureFormats		? ObjectFormatCode Array
	PTP_UINT32(27),							//10 ImageFormats		? ObjectFormatCode Array
//	 PTP_UINT16(PTP_ObjectFormat_Undefined),		//		0x3000		//Undefined non-image object						
	 PTP_UINT16(PTP_ObjectFormat_Association),		// 1		0x3001		//Association (e.g. folder)						
	 PTP_UINT16(PTP_ObjectFormat_Script),			// 2		0x3002		//Device-model-specific script						
	 PTP_UINT16(PTP_ObjectFormat_Executable),		// 3		0x3003		//Device-model-specific binary executable				
	 PTP_UINT16(PTP_ObjectFormat_Text),			// 4		0x3004		//Text file								
	 PTP_UINT16(PTP_ObjectFormat_HTML),			// 5		0x3005		//HyperText Markup Language file (text)					
	 PTP_UINT16(PTP_ObjectFormat_DPOF),			// 6		0x3006		//Digital Print Order Format file (text)				
	 PTP_UINT16(PTP_ObjectFormat_AIFF),			// 7		0x3007		//Audio clip								
	 PTP_UINT16(PTP_ObjectFormat_WAV),			// 8		0x3008		//Audio clip								
	 PTP_UINT16(PTP_ObjectFormat_MP3),			// 9		0x3009		//Audio clip								
	 PTP_UINT16(PTP_ObjectFormat_AVI),			//10		0x300A		//Video clip								
	 PTP_UINT16(PTP_ObjectFormat_MPEG),			//11		0x300B		//Video clip								
	 PTP_UINT16(PTP_ObjectFormat_ASF),			//12		0x300C		//Microsoft Advanced Streaming Format (video)				
//	 PTP_UINT16(PTP_ObjectFormat_Undefined),		//		0x3800		//Unknown image object							
	 PTP_UINT16(PTP_ObjectFormat_EXIF_JPEG),		//13		0x3801		//Exchangeable File Format, JEIDA standard				
	 PTP_UINT16(PTP_ObjectFormat_TIFF_EP),			//14		0x3802		//Tag Image File Format for Electronic Photography			`ĂIuWFNgtH[}bgASė񋓂ĂƂɂB
	 PTP_UINT16(PTP_ObjectFormat_FlashPix),			//15		0x3803		//Structured Storage Image Format					ɗ񋓂ĂȂ`́AOSŔfāARs[sɂ邩B
	 PTP_UINT16(PTP_ObjectFormat_BMP),			//16		0x3804		//Microsoft Windows Bitmap file						
	 PTP_UINT16(PTP_ObjectFormat_CIFF),			//17		0x3805		//Canon Camera Image File Format					
//	 PTP_UINT16(PTP_ObjectFormat_Undefined),		//		0x3806		//Reserved								
	 PTP_UINT16(PTP_ObjectFormat_GIF),			//18		0x3807		//Graphics Interchange Format						
	 PTP_UINT16(PTP_ObjectFormat_JFIF),			//19		0x3808		//JPEG File Interchange Format						
	 PTP_UINT16(PTP_ObjectFormat_PCD),			//20		0x3809		//PhotoCD Image Pac							
	 PTP_UINT16(PTP_ObjectFormat_PICT),			//21		0x380A		//Quickdraw Image Format						
	 PTP_UINT16(PTP_ObjectFormat_PNG),			//22		0x380B		//Portable Network Graphics						
//	 PTP_UINT16(PTP_ObjectFormat_Undefined),		//		0x380C		//Reserved								
	 PTP_UINT16(PTP_ObjectFormat_TIFF),			//23		0x380D		//Tag Image File Format							
	 PTP_UINT16(PTP_ObjectFormat_TIFF_IT),			//24		0x380E		//Tag Image File Format for Information Technology (graphic arts)	
	 PTP_UINT16(PTP_ObjectFormat_JP2),			//25		0x380F		//JPEG2000 Baseline File Format						
	 PTP_UINT16(PTP_ObjectFormat_JPX),			//26		0x3810		//JPEG2000 Extended File Format						
//	 PTP_UINT16(PTP_ObjectFormat_Undefined),		//		0x3000/*+?*/	//Reserved for future use	All other codes with MSN of 0011	
	 PTP_UINT16(PTP_ObjectFormat_VendorDefined),		//27		0xB000/*+?*/	//Vendor-Defined		All other codes with MSN of 1011	
	11,								//11 Manufacturer		? String
	 PTP_UINT16('P'),// 1
	 PTP_UINT16('i'),// 2
	 PTP_UINT16('e'),// 3
	 PTP_UINT16('c'),// 4
	 PTP_UINT16('e'),// 5
	 PTP_UINT16(' '),// 6
	 PTP_UINT16('L'),// 7
	 PTP_UINT16('a'),// 8
	 PTP_UINT16('b'),// 9
	 PTP_UINT16('.'),//10
	 PTP_UINT16( 0 ),//11
	17,								//12 Model			? String
	 PTP_UINT16('P'),// 1
	 PTP_UINT16('/'),// 2
	 PTP_UINT16('E'),// 3
	 PTP_UINT16('C'),// 4
	 PTP_UINT16('E'),// 5
	 PTP_UINT16(' '),// 6
	 PTP_UINT16('U'),// 7
	 PTP_UINT16('S'),// 8
	 PTP_UINT16('B'),// 9
	 PTP_UINT16(' '),//10
	 PTP_UINT16('d'),//11
	 PTP_UINT16('e'),//12
	 PTP_UINT16('v'),//13
	 PTP_UINT16('i'),//14
	 PTP_UINT16('c'),//15
	 PTP_UINT16('e'),//16
	 PTP_UINT16( 0 ),//17
	0,								//13 DeviceVersion		? String
	0,								//14 SerialNumber		? String
};
/*--------------------------------------------------------------------------*/
//If a physical store does not contain any logical stores, the LogicalStorageID should be set to 0x0000, and it shall be assumed that the store does not contain any data, nor can it receive any data.
//StorageIDs of 0x00000000 and 0xFFFFFFFF are not used to refer to real stores, but have context-specific meanings, such as gall-stores,h or gthe default store.h
#define StorageID	(1|1<<16)	//PhysicalStorageID=1,LogicalStorageID=1
/*--------------------------------------------------------------------------*/
/* StorageID Array */
static const uint8_t StorageIDArray[] = {
	PTP_UINT32(1),
	 PTP_UINT32(StorageID),// 0
};
/*--------------------------------------------------------------------------*/
/* StorageInfo Dataset */
static const uint8_t StorageInfoDataset[] = {
	PTP_UINT16(PTP_StorageType_FixedRAM),				// 1 StorageType		2 UINT16
	PTP_UINT16(PTP_FilesystemType_GenericFlat),			// 2 FilesystemType		2 UINT16
	PTP_UINT16(PTP_AccessCapability_ReadWrite),			// 3 AccessCapability		2 UINT16
	PTP_UINT64(-1,-1),						// 4 MaxCapacity		8 UINT64	If this field is unused, it should report 0xFFFFFFFF.
	PTP_UINT64(-1,-1),						// 5 FreeSpaceInBytes		8 UINT64	If this value is not useful for the device, it may set this field to 0xFFFFFFFF and rely upon the FreeSpaceInImages field instead.
	PTP_UINT32(-1),							// 6 FreeSpaceInImages		4 UINT32	If the device does not implement this capability, this field should be set to 0xFFFFFFFF. This field may be used for devices that do not report FreeSpaceInBytes, or the two fields may be used in combination.
	0,								// 7 StorageDescription		? String
	0,								// 8 VolumeLabel		? String
};
/****************************************************************************
 *	
 ****************************************************************************/
void PTP_Start(const char* pProtectedFilenames) {
	/* mɏIĂB */
	PTP_Stop();
	/* veNgt@C̃Xgi[B */
	stPTP.pProtectedFilenames = pProtectedFilenames ? /*GCCg@\*/ : "";
	/* USB̃tbNJnB */
	usb_start(&usb_inf);
}
/*--------------------------------------------------------------------------*/
void PTP_Stop() {
	/* USB̃tbNIB */
	usb_stop();
	/* \̃NAB */
	memset(&stPTP, 0, sizeof stPTP);
}
/****************************************************************************
 *	
 ****************************************************************************/
static void get_descriptor(const DEVICE_REQUEST* req, const unsigned char* dat, int len) {
	int type  = HIBYTE(req->wValue);
	int index = LOBYTE(req->wValue);
	switch(type) {
	case DEVICE_DESCRIPTOR_TYPE:
		usb_ep0_tx(&device_desc, sizeof(DEVICE_DESCRIPTOR));
		return;
	case CONFIGURATION_DESCRIPTOR_TYPE:
		if(index == 0) {
			static uint8_t buf[sizeof(CONFIGURATION_DESCRIPTOR) +	/* config_desc */
					   sizeof(INTERFACE_DESCRIPTOR) +	/* interface_desc */
					   sizeof(ENDPOINT_DESCRIPTOR) * 3];	/* endpoint_desc_ep2in,endpoint_desc_ep2out,endpoint_desc_ep1in */
			uint8_t* ptr =  memcpy(buf, &config_desc,          sizeof(CONFIGURATION_DESCRIPTOR))  + sizeof(CONFIGURATION_DESCRIPTOR);
			         ptr =  memcpy(ptr, &interface_desc,       sizeof(INTERFACE_DESCRIPTOR)    )  + sizeof(INTERFACE_DESCRIPTOR);
			         ptr =  memcpy(ptr, &endpoint_desc_ep2in,  sizeof(ENDPOINT_DESCRIPTOR)     )  + sizeof(ENDPOINT_DESCRIPTOR);
			         ptr =  memcpy(ptr, &endpoint_desc_ep2out, sizeof(ENDPOINT_DESCRIPTOR)     )  + sizeof(ENDPOINT_DESCRIPTOR);
			       /*ptr =*/memcpy(ptr, &endpoint_desc_ep1in,  sizeof(ENDPOINT_DESCRIPTOR)     )/*+ sizeof(ENDPOINT_DESCRIPTOR)*/;
			usb_ep0_tx(buf, sizeof buf);
			return;
		}
		break;
	case STRING_DESCRIPTOR_TYPE:
		if(index < ARRAY_SIZE(string_desc)) {
			usb_ep0_tx(string_desc[index], string_desc[index][0]);
			return;
		}
		break;
	}
	usb_ep0_stall(); /* Ή */
}
/*--------------------------------------------------------------------------*/
static void set_configuration(const DEVICE_REQUEST* req, const unsigned char* dat, int len) {
	D12_SetEndpointEnable(req->wValue == 1);
	usb_ep0_tx(NULL, 0); /* ACK */
}
/*--------------------------------------------------------------------------*/
static void set_interface(const DEVICE_REQUEST* req, const unsigned char* dat, int len) {
	usb_ep0_tx(NULL, 0); /* ACK */
}
/*--------------------------------------------------------------------------*/
static void class_request(const DEVICE_REQUEST* req, const unsigned char* dat, int len) {
	switch(req->bRequest) {
	/*--------------------------------------------------------------------------*/
//	case 0x64://Cancel Request
//	case 0x65://Get Extended Event Data
//	case 0x66://Device Reset Request
	/*--------------------------------------------------------------------------*/
	case 0x67://Get Device Status Request
		if((D12_ReadEndpointStatus(EP2OUT) | D12_ReadEndpointStatus(EP2IN)) & D12_ENDPOINTSTALLED) {
			static const uint8_t DeviceStatusData[]={
				PTP_UINT16(2+2+4+4),					//wLength	This field specifies the total length of the status data
				PTP_UINT16(PTP_ResponseCode_TransactionCancelled),	//Code		The PIMA15740 Response Code or Vendor Code: 0x2001 Status OK 0x2019 Device Busy 0x201F Transaction Cancelled
				PTP_UINT32(0x02),	/* EP2OUT */			//Parameter 1	The format and meaning of the parameter depends on the Status Code. For device initiated cancels this parameter contains an endpoint number.	dl(uPHOTOGRAPHIC AND IMAGING MANUFACTURERS ASSOCIATION, INC. PIMA 15740:2000vpima15740-2000.pdf)ɂ͖LĂȂłA
				PTP_UINT32(0x82),	/* EP2IN */			//Parameter N	The format and meaning of the parameter depends on the Status Code. For device initiated cancels this parameter contains an endpoint number.	ParameterUINT32złBGh|Cgԍ̔zȂ̂UINT8z񂩂ƎvłAUINT32z񂪐łB(Windows7̋萄)
			};
			usb_ep0_tx(DeviceStatusData, sizeof DeviceStatusData);
		} else {
			static const uint8_t DeviceStatusData[]={
				PTP_UINT16(2+2),					//wLength	This field specifies the total length of the status data
				PTP_UINT16(PTP_ResponseCode_OK),			//Code		The PIMA15740 Response Code or Vendor Code: 0x2001 Status OK 0x2019 Device Busy 0x201F Transaction Cancelled
			};
			usb_ep0_tx(DeviceStatusData, sizeof DeviceStatusData);
		}
		return;
	/*--------------------------------------------------------------------------*/
	}
	/* The Device may support class-specific requests.
	 * The device shall return STALL if an unrecognized or unsupported device-specific request is received. */
	usb_ep0_stall(); /* Ή */
}
/*--------------------------------------------------------------------------*/
static void ep2_rxdone() {
	union { uint8_t buf[EP2_NONISO_PACKET_SIZE]; PTP_GenericContainerStructure Container; } u;
	int len;
	for(;;) {
		/* Mf[^ǂݏoBMf[^΋ABMf[^ǂݏoAǂݏoMf[^NAB
		 * - D12_ReadBuffer()(-1)ꍇ́AD12_ClearBuffer()Ăł͂ȂB
		 *   D12_ReadBuffer()`D12_ClearBuffer()̊ԂɐVf[^ꍇAǂݏoOɃNAoOɂȂB */
		memset(&u, 0, sizeof u);
		if((len = D12_ReadBuffer(EP2OUT, &u, sizeof u)) == -1) { return; } /* ܂ */
		         D12_ClearBuffer(EP2OUT);
		/* R}htF[YȂ΁c */
		if(stPTP.Phase == PTP_Phase_Command) {
			/* ȉ̌ɕsiȂ΁AG[ƌȂăXg[B
			 * - R}htF[Y̎Mf[^́AyC[hTCY(64Byte)ł͂B
			 *   (Reiwb_(12Byte)+p[^~0`5(0`20Byte)=12`32Byte64Byte) */
			if((len == EP2_NONISO_PACKET_SIZE) ||					//yC[hTCYłȂ
			   (u.Container.ContainerLength != len) ||				//ReiƈvĂȂ
			   (u.Container.ContainerType != PTP_ContainerType_CommandBlock)) {	//R}hubNłȂ
				break; } /* stall */
			/* Iy[VNGXgf[^Zbg\zB */
			memset(&stPTP.OperationRequestDataset, 0, sizeof stPTP.OperationRequestDataset);
			stPTP.OperationRequestDataset.OperationCode = u.Container.Code;
			stPTP.TransactionID                         = u.Container.TransactionID;
			if((len -= offsetof(PTP_GenericContainerStructure, Payload)) < 0) { len = 0; }
			if(len > sizeof stPTP.OperationRequestDataset.Parameter) { len = sizeof stPTP.OperationRequestDataset.Parameter; }
			memcpy(stPTP.OperationRequestDataset.Parameter, u.Container.Payload, len);
			/* R}hBG[Ȃ΁AXg[B */
			if(process_command()) { break; } /* stall */
			/* f[^CtF[Y,,X|XtF[YJnBf[^AEgtF[YȂ΁Aep2_txsub()B */
			if(ep2_txsub()) { break; } /* stall */
		/* f[^AEgtF[YȂ΁c */
		} else if(stPTP.Phase == PTP_Phase_DataOut) {
			/* Ō̃pPbgł邩𔻒肵ĂB
			 * - f[^AEgtF[Y̎Mf[^́AyC[hTCỸpPbgŏI[B
			 *   Reiłfł邪AIɃyC[hTCỸpPbgŏI[̂APTP̎dlłB
			 *   The Data phase ends when the number of bytes transferred equals the number of bytes specified in the first four bytes of the Data Block that coincide with a short or a NULL packet.
			 *   A short packet or a NULL packet indicates the end of the end of a Data phase.
			 *   If the number of bytes specified in the first four bytes of the Data Block are an integral multiple of the wMaxPacketSize field of the Endpoint Descriptor the Data phase will end in a NULL packet. */
			int end = (len != EP2_NONISO_PACKET_SIZE);
			/* ŏ̃pPbgȂ΁c */
			if(!stPTP.DataPos) {
				/* f[^ubNłȂ΁AG[ƌȂăXg[B */
				if(u.Container.ContainerType != PTP_ContainerType_DataBlock) { break; } /* stall */
				/* f[^擾Bf[^0Ȃ΁AG[ƌȂăXg[B */
				stPTP.DataLen = u.Container.ContainerLength - offsetof(PTP_GenericContainerStructure, Payload);
				if(stPTP.DataLen < 0) { break; } /* stall */
				/* MTCYAReiwb_̃TCYB */
				if((len -= offsetof(PTP_GenericContainerStructure, Payload)) < 0) { len = 0; }
				/* f[^obt@̐擪ւ炷B */
				memmove(&u, u.Container.Payload, len);
			}
			/* Ō̃pPbgMf[^sv,,f[^ʒuf[^𒴉߂AG[ƌȂB */
			if((end && ((stPTP.DataPos + len) != stPTP.DataLen)) ||
			           ((stPTP.DataPos + len) >  stPTP.DataLen)) { break; } /* stall */
			/* f[^B */
			if(len && process_dataout(u.buf, len)) { break; } /* stall */
			/* f[^ʒuZB */
			stPTP.DataPos += len;
			/* Ō̃pPbgȂ΁c */
			if(end) {
				/* X|XtF[YֈڍsB */
				stPTP.Phase = PTP_Phase_Response;
				/* X|XtF[YJnB */
				if(ep2_txsub()) { break; } /* stall */
			}
		/* ȊÕtF[YȂ΁c */
		} else {
			/* ȊÕtF[Ył͎MȂ͂łBMꍇ́AG[ƌȂăXg[B */
			break; /* stall */
		}
		/* MpB
		 * _uobt@̂ʂɎMf[^cĂ邩,,̏ɐVf[^\L邩B */
	}
	/* STALL Bulk-in & Bulk-out Endpoints */
	D12_SetEndpointStatus(EP2OUT, 1);
	D12_SetEndpointStatus(EP2IN , 1);
	/* R}htF[Y֖߂B */
	stPTP.Phase = PTP_Phase_Command;
}
/*--------------------------------------------------------------------------*/
static void ep2_txdone() {
	if(ep2_txsub()) {
		/* STALL Bulk-in & Bulk-out Endpoints */
		D12_SetEndpointStatus(EP2OUT, 1);
		D12_SetEndpointStatus(EP2IN , 1);
		/* R}htF[Y֖߂B */
		stPTP.Phase = PTP_Phase_Command;
	}
}
/*--------------------------------------------------------------------------*/
static int ep2_txsub() {
	union { uint8_t buf[EP2_NONISO_PACKET_SIZE]; PTP_GenericContainerStructure Container; } u;
	int len;
	for(;;) {
		/* f[^CtF[YȂ΁c */
		if(stPTP.Phase == PTP_Phase_DataIn) {
			int end;
			/* ŏ̃pPbgȂ΁c */
			if(!stPTP.DataPos) {
				/* Rei\zB */
				memset(&u, 0, sizeof u);
				u.Container.ContainerLength = offsetof(PTP_GenericContainerStructure, Payload) + stPTP.DataLen;
				u.Container.ContainerType   = PTP_ContainerType_DataBlock;
				u.Container.Code            = stPTP.OperationRequestDataset.OperationCode;	//The Data Block will use the OperationCode from the Command Block
				u.Container.TransactionID   = stPTP.TransactionID;
				/* ŏ̃pPbgɓ镪́Af[^擾B */
				len = (stPTP.DataLen - stPTP.DataPos);
				if(len > (EP2_NONISO_PACKET_SIZE - offsetof(PTP_GenericContainerStructure, Payload))) {
				   len = (EP2_NONISO_PACKET_SIZE - offsetof(PTP_GenericContainerStructure, Payload)); }
				if(len && process_datain(u.Container.Payload, len)) { return -1; }
			/* ŏ̃pPbgłȂ΁c */
			} else {
				/* f[^擾B */
				len = (stPTP.DataLen - stPTP.DataPos);
				if(len > EP2_NONISO_PACKET_SIZE) { len = EP2_NONISO_PACKET_SIZE; }
				if(len && process_datain(u.buf, len)) { return -1; }
			}
			/* ŏ̃pPbgȂ΁AReiwb_̃TCYZAMTCYƂB */
			if(!stPTP.DataPos) { len += offsetof(PTP_GenericContainerStructure, Payload); }
			/* pPbgށBpPbg񂾂ApPbgoJnB
			 * - pPbg߂Ȃ΁ApPbg𑗏oɁA(0)ԂB
			 *   ̌őMobt@󂢂Aep2_txdone()͂łB
			 *   ep2_txdone()AēxAsB */
			if((len = D12_WriteBuffer(EP2IN, &u, len)) == -1) { return 0; }
			       D12_ValidateBuffer(EP2IN);
			/* Ō̃pPbgł邩𔻒肵ĂB
			 * - f[^CtF[Y̑Mf[^́AyC[hTCỸpPbgŏI[B
			 *   Reiłfł邪AIɃyC[hTCỸpPbgŏI[̂APTP̎dlłB
			 *   The Data phase ends when the number of bytes transferred equals the number of bytes specified in the first four bytes of the Data Block that coincide with a short or a NULL packet.
			 *   A short packet or a NULL packet indicates the end of the end of a Data phase.
			 *   If the number of bytes specified in the first four bytes of the Data Block are an integral multiple of the wMaxPacketSize field of the Endpoint Descriptor the Data phase will end in a NULL packet. */
			end = (len != EP2_NONISO_PACKET_SIZE);
			/* ŏ̃pPbgȂ΁AMTCYAReiwb_̃TCYB */
			if(!stPTP.DataPos) { if((len -= offsetof(PTP_GenericContainerStructure, Payload)) < 0) { len = 0; } }
			/* f[^ʒuZB */
			stPTP.DataPos += len;
			/* Ō̃pPbgȂ΁AX|XtF[YֈڍsB */
			if(end) { stPTP.Phase = PTP_Phase_Response; }
		/* X|XtF[YȂ΁c */
		} else if(stPTP.Phase == PTP_Phase_Response) {
			/* Rei\zB
			 * - X|XtF[Y̑Mf[^́AyC[hTCY(64Byte)łB
			 *   (Reiwb_(12Byte)+p[^~5(20Byte)=32Byte64Byte) */
			memset(&u, 0, sizeof u);
			u.Container.ContainerLength = offsetof(PTP_GenericContainerStructure, Payload) + sizeof stPTP.ResponseDataset.Parameter;
			u.Container.ContainerType   = PTP_ContainerType_ResponseBlock;
			u.Container.Code            = stPTP.ResponseDataset.ResponseCode;
			u.Container.TransactionID   = stPTP.TransactionID;
			memcpy(u.Container.Payload, stPTP.ResponseDataset.Parameter, sizeof stPTP.ResponseDataset.Parameter);
			/* pPbgށBpPbg񂾂ApPbgoJnB
			 * - pPbg߂Ȃ΁ApPbg𑗏oɁA(0)ԂB
			 *   ̌őMobt@󂢂Aep2_txdone()͂łB
			 *   ep2_txdone()AēxAsB */
			if((len = D12_WriteBuffer(EP2IN, &u, u.Container.ContainerLength)) == -1) { return 0; }
			       D12_ValidateBuffer(EP2IN);
			/* R}htF[Y֖߂B */
			stPTP.Phase = PTP_Phase_Command;
		/* ȊÕtF[YȂ΁c */
		} else {
			/* ȊÕtF[Ył͑MȂBMɁA(0)ԂB */
			return 0;
		}
		/* MpB
		 * _uobt@̂ʂ󂢂Ă邩,,̏Ƀf[^oobt@󂢂\L邩B */
	}
}
/*--------------------------------------------------------------------------*/
static void put_string(void* _dst, const char* src) {
	char* len = _dst;	//NumChars
	char* dst = _dst + 1;	//StringChars
	*len = 0;
	if(*src) {		//(NumChars=0)̏ꍇAI[̃ki[ȂB
		int c;
		do {
			*dst++ = c = *src++;
			*dst++ = 0;
			(*len)++;
		} while(c);
	}
}
/*--------------------------------------------------------------------------*/
static void get_string(char* dst, const void* _src) {
	const uint8_t* src = _src;
	int len = *src++;	//NumChars
	while(len--) {
		*dst++ = *src;
		          src += 2;
	}
	*dst = 0;		//(NumChars=0)̏ꍇAI[̃ktB
}
/*--------------------------------------------------------------------------*/
static int is_protected_filename(const char* name) {
	const char* ptr = stPTP.pProtectedFilenames;
	/* veNgt@C̃Xg̏I[ɒB܂Łc */
	while(*ptr) {
		/* veNgt@CɈvA1ԂB */
		if(!strcasecmp(ptr, name)) { return 1; }
		/* veNgt@CA֐i߂B */
		ptr += (strlen(ptr) + 1);
	}
	/* veNgt@CɈvȂ΁A0ԂB */
	return 0;
}
/*--------------------------------------------------------------------------*/
static int process_command() {
	/* R}htF[Y̎ɍsׂtF[Y擾B */
	int i = stPTP.OperationRequestDataset.OperationCode - PTP_OperationCode_GetDeviceInfo;
	if((unsigned)i >= ARRAY_SIZE(TBL_DataDirection)) { return -1; }
	stPTP.Phase = TBL_DataDirection[i];
	/* ȍ~̃R[hʂ팸邽߂ɁAʓIȒlݒ肵ĂB */
	memset(&stPTP.ResponseDataset, 0, sizeof stPTP.ResponseDataset);
	stPTP.ResponseDataset.ResponseCode = PTP_ResponseCode_OK;
	stPTP.DataLen = stPTP.DataPos = 0, stPTP.DataPtr = NULL;
	/* Iy[VR[hɂāc */
	switch(stPTP.OperationRequestDataset.OperationCode) {
	/*--------------------------------------------------------------------------*/
	case PTP_OperationCode_GetDeviceInfo:
		stPTP.DataLen = sizeof DeviceInfoDataset;
		stPTP.DataPtr = (void*)DeviceInfoDataset;
		return 0;
	/*--------------------------------------------------------------------------*/
	case PTP_OperationCode_OpenSession:
		stPTP.SessionID = stPTP.OperationRequestDataset.Parameter[0];
		return 0;
	/*--------------------------------------------------------------------------*/
	case PTP_OperationCode_CloseSession:
		stPTP.SessionID = 0;
		return 0;
	/*--------------------------------------------------------------------------*/
	case PTP_OperationCode_GetStorageIDs:
		stPTP.DataLen = sizeof StorageIDArray;
		stPTP.DataPtr = (void*)StorageIDArray;
		return 0;
	/*--------------------------------------------------------------------------*/
	case PTP_OperationCode_GetStorageInfo:
		{
			/* t@CVXe̓v擾B */
			struct statvfs st;
			if(statvfs("/", &st)) { return -1; }
			/* Xg[Wf[^ZbgRs[AMaxCapacity,FreeSpaceInBytesi[B */
			// 1 StorageType		+ 0,2 UINT16
			// 2 FilesystemType		+ 2,2 UINT16
			// 3 AccessCapability		+ 4,2 UINT16
			// 4 MaxCapacity		+ 6,8 UINT64
			// 5 FreeSpaceInBytes		+14,8 UINT64
			// 6 FreeSpaceInImages		+22,4 UINT32
			// 7 StorageDescription		+26,? String
			// 8 VolumeLabel		+ ?,? String
			memcpy(stPTP.WorkBuf, StorageInfoDataset, sizeof StorageInfoDataset);
			PUT_LEWORD(stPTP.WorkBuf +  6, st.f_frsize * st.f_blocks);	//MaxCapacity
			PUT_LEWORD(stPTP.WorkBuf + 10, 0);				//
			PUT_LEWORD(stPTP.WorkBuf + 14, st.f_frsize * st.f_bfree);	//FreeSpaceInBytes
			PUT_LEWORD(stPTP.WorkBuf + 18, 0);				//
			/* Xg[Wf[^ZbgAf[^CtF[Y̑Mf[^ƂB */
			stPTP.DataLen = sizeof StorageInfoDataset;
			stPTP.DataPtr = stPTP.WorkBuf;
		}
		return 0;
	/*--------------------------------------------------------------------------*/
	case PTP_OperationCode_GetNumObjects:
	case PTP_OperationCode_GetObjectHandles:
		{
			/* IuWFNgnhz쐬B
			 * readdir()g킸ɁAfBNgGg𒼐ڑBIuWFNgnh𓾂邽߂łB */
			const pffsMASTERBLOCK* msb = (pffsMASTERBLOCK*)pceSystemGetInfo()->pffs_top;
			const DIRECTORY *pdir = msb->dir, *pdir_end = msb->dir + MAXDIR;
			uint32_t* ObjectHandleArray = (uint32_t*)stPTP.WorkBuf;
			ObjectHandleArray[0/*NumElements*/] = 0;
			do {
				/* gp̃fBNgGgȂ΁c(\usr\PIECE\sysdev\pcekn\file.c Search() Q) */
				if((pdir->name[0] == 0) || (pdir->name[0] == -1)) { continue; }	//
				/* z̗vf𑝂₵AIuWFNgnhi[B
				 * fBNgGgւ̃|C^AIuWFNgnhƂĊi[邱ƂɂB */
				ObjectHandleArray[++ObjectHandleArray[0/*NumElements*/]] = (uint32_t)pdir;
			} while(++pdir < pdir_end);
			switch(stPTP.OperationRequestDataset.OperationCode) {
			/*--------------------------------------------------------------------------*/
			case PTP_OperationCode_GetNumObjects:
				/* IuWFNgAX|X̃p[^Ɋi[B */
				stPTP.ResponseDataset.Parameter[0] = ObjectHandleArray[0/*NumElements*/];
				break;
			/*--------------------------------------------------------------------------*/
			case PTP_OperationCode_GetObjectHandles:
				/* IuWFNgnhzAf[^CtF[Y̑Mf[^ƂB */
				stPTP.DataLen = sizeof(uint32_t) * (1 + ObjectHandleArray[0/*NumElements*/]);
				stPTP.DataPtr = ObjectHandleArray;
				break;
			/*--------------------------------------------------------------------------*/
			default: DIE();
			}
		}
		return 0;
	/*--------------------------------------------------------------------------*/
	case PTP_OperationCode_GetObjectInfo:
		{
			const DIRECTORY* pdir = (DIRECTORY*)stPTP.OperationRequestDataset.Parameter[0];
			struct stat st;
			if(stat(pdir->name, &st)) {	//IuWFNgnhs
				stPTP.ResponseDataset.ResponseCode = PTP_ResponseCode_InvalidObjectHandle;
				return 0;
			}
			/* IuWFNgf[^Zbg\zB */
			// 1 StorageID			+ 0,4 StorageID
			// 2 ObjectFormat		+ 4,2 ObjectFormatCode		WindowsXpŎƂA@VendorDefined*ȊO*ɂƁAGNXv[ɊgqςĂ܂BAVendorDefinedɂƁAGNXv[Ń_uNbNăvr[悤ƂƃGNXv[sIB@̕AdȂ̂ŁAVendorDefinedɂ邱ƂɂB
			// 3 ProtectionStatus		+ 6,2 UINT16			Windows7łƂA@̖͔ȂBA_uNbNɂǂ̃r[AN邩ɂ͉e悤BA̖ȂB_ƂẮAWindows7łAVendorDefinedɂĖȂB
			// 4 ObjectCompressedSize	+ 8,4 UINT32			WindowsXpŁAʂ̖肪BgqقȂ̃t@CLƁAWindowsXp̃GNXv[ł́AЕ̃t@C\Ă܂BWindowsXṕAgqt@CObjectFormat̑gŁAĂ̂ł͂ȂƎvB@B
			// 5 ThumbFormat		+12,2 ObjectFormatCode		@@ɈƂɁAWindowsXp̃GNXv[oRŁÃt@C𗼕폜悤ƂƁAP/ECEnOAbvBǂ̂悤ȃNGXgŁAP/ECEnOAbvĂ̂A܂𒲍łĂȂB
			// 6 ThumbCompressedSize	+14,4 UINT32
			// 7 ThumbPixWidth		+18,4 UINT32
			// 8 ThumbPixHeight		+22,4 UINT32
			// 9 ImagePixWidth		+26,4 UINT32
			//10 ImagePixHeight		+30,4 UINT32
			//11 ImageBitDepth		+34,4 UINT32
			//12 ParentObject		+38,4 ObjectHandle
			//13 AssociationType		+42,2 AssociationCode
			//14 AssociationDesc		+44,4 AssociationDesc
			//15 SequenceNumber		+48,4 UINT32
			//16 Filename			+52,? String		1+((8+1+3)+1)*2=27Byte(Max)
			//17 CaptureDate		+ ?,? DateTime String	1
			//18 ModificationDate		+ ?,? DateTime String	1
			//19 Keywords			+ ?,? String		1
			memset(stPTP.WorkBuf, 0, sizeof stPTP.WorkBuf);
			PUT_LEWORD(stPTP.WorkBuf +  0, StorageID);				//StorageID
			PUT_LEHALF(stPTP.WorkBuf +  4, PTP_ObjectFormat_VendorDefined);		//ObjectFormat		LRgQ
		//sv	PUT_LEHALF(stPTP.WorkBuf +  6, PTP_ProtectionStatus_NoProtection);	//ProtectionStatus	(PTP_ProtectionStatus_NoProtection=0)Ȃ̂ŕsv
			PUT_LEWORD(stPTP.WorkBuf +  8, st.st_size);				//ObjectCompressedSize
			put_string(stPTP.WorkBuf + 52, pdir->name);				//Filename
			if(is_protected_filename(pdir->name)) {//veNgt@C?
				PUT_LEHALF(stPTP.WorkBuf +  6, PTP_ProtectionStatus_ReadOnly);	//ProtectionStatus
			}
			/* IuWFNgf[^ZbgAf[^CtF[Y̑Mf[^ƂB */
			stPTP.DataLen = 52 + (1 + stPTP.WorkBuf[52] * 2) + 3;
			stPTP.DataPtr = stPTP.WorkBuf;
		}
		return 0;
	/*--------------------------------------------------------------------------*/
	case PTP_OperationCode_GetObject:
		{
			const DIRECTORY* pdir = (DIRECTORY*)stPTP.OperationRequestDataset.Parameter[0];
			struct stat st;
			if(stat(pdir->name, &st)) {	//IuWFNgnhs
				stPTP.ResponseDataset.ResponseCode = PTP_ResponseCode_InvalidObjectHandle;
				return 0;
			}
			/* f[^CtF[ÝAMf[^ݒ肷B */
			stPTP.DataLen = st.st_size;
		}
		return 0;
	/*--------------------------------------------------------------------------*/
//	case PTP_OperationCode_GetThumb:
	/*--------------------------------------------------------------------------*/
	case PTP_OperationCode_DeleteObject:
		{
			int retval;
			const DIRECTORY* pdir = (DIRECTORY*)stPTP.OperationRequestDataset.Parameter[0];
			if(is_protected_filename(pdir->name)) {//veNgt@C?
				stPTP.ResponseDataset.ResponseCode = PTP_ResponseCode_ObjectWriteProtected;
				return 0;
			}
			/* t@C폜B */
			memcpy(stPTP.save_FRAM4, FRAM4_START, FRAM4_SIZE);	//FRAM4ޔ
			retval = pceFileDelete(pdir->name);			//FRAM4j󂳂
			memcpy(FRAM4_START, stPTP.save_FRAM4, FRAM4_SIZE);	//FRAM4𕜌
			if(retval) {	//IuWFNgnhs
				stPTP.ResponseDataset.ResponseCode = PTP_ResponseCode_InvalidObjectHandle;
				return 0;
			}
		}
		return 0;
	/*--------------------------------------------------------------------------*/
	case PTP_OperationCode_SendObjectInfo:
	case PTP_OperationCode_SendObject:
		/** no job **/
		return 0;
	/*--------------------------------------------------------------------------*/
//	case PTP_OperationCode_InitiateCapture:
//	case PTP_OperationCode_FormatStore:
//	case PTP_OperationCode_ResetDevice:
//	case PTP_OperationCode_SelfTest:
//	case PTP_OperationCode_SetObjectProtection:
//	case PTP_OperationCode_PowerDown:
//	case PTP_OperationCode_GetDevicePropDesc:
//	case PTP_OperationCode_GetDevicePropValue:
//	case PTP_OperationCode_SetDevicePropValue:
//	case PTP_OperationCode_ResetDevicePropValue:
//	case PTP_OperationCode_TerminateOpenCapture:
//	case PTP_OperationCode_MoveObject:
//	case PTP_OperationCode_CopyObject:
//	case PTP_OperationCode_GetPartialObject:
//	case PTP_OperationCode_InitiateOpenCapture:
	/*--------------------------------------------------------------------------*/
	}
	/* ΁AG[ƂB */
	return -1;
}
/*--------------------------------------------------------------------------*/
static int process_dataout(const void* buf, int len) {
	/* f[^|C^ݒ肳ĂAf[^i[B */
	if(stPTP.DataPtr) { memcpy(stPTP.DataPtr + stPTP.DataPos, buf, len); }
	/* Iy[VR[hɂāc */
	switch(stPTP.OperationRequestDataset.OperationCode) {
	/*--------------------------------------------------------------------------*/
	case PTP_OperationCode_SendObjectInfo:
		/* ŏ̃pPbgȂ΁c */
		if(!stPTP.DataPos) {
			/* IuWFNgf[^ZbgA[Nobt@Ɋi[ł邱ƂmFB */
			if(stPTP.DataLen > sizeof stPTP.WorkBuf) { return -1; }	//s
			stPTP.DataPtr = stPTP.WorkBuf;
			/* ŏ̃pPbgi[B񂩂́A֐̐擪̋ʏŊi[B */
			memcpy(stPTP.DataPtr + stPTP.DataPos, buf, len);
		}
		/* Ō̃pPbgȂ΁c */
		if(stPTP.DataPos + len == stPTP.DataLen) {
			char name[(8+1+3)+1];
			{
				int retval;
				// 1 StorageID			+ 0,4 StorageID
				// 2 ObjectFormat		+ 4,2 ObjectFormatCode
				// 3 ProtectionStatus		+ 6,2 UINT16
				// 4 ObjectCompressedSize	+ 8,4 UINT32
				// 5 ThumbFormat		+12,2 ObjectFormatCode
				// 6 ThumbCompressedSize	+14,4 UINT32
				// 7 ThumbPixWidth		+18,4 UINT32
				// 8 ThumbPixHeight		+22,4 UINT32
				// 9 ImagePixWidth		+26,4 UINT32
				//10 ImagePixHeight		+30,4 UINT32
				//11 ImageBitDepth		+34,4 UINT32
				//12 ParentObject		+38,4 ObjectHandle
				//13 AssociationType		+42,2 AssociationCode
				//14 AssociationDesc		+44,4 AssociationDesc
				//15 SequenceNumber		+48,4 UINT32
				//16 Filename			+52,? String
				//17 CaptureDate		+ ?,? DateTime String
				//18 ModificationDate		+ ?,? DateTime String
				//19 Keywords			+ ?,? String
				uint32_t size = LEWORD(stPTP.WorkBuf +  8);	//ObjectCompressedSize
				uint8_t* pstr =        stPTP.WorkBuf + 52 ;	//Filename
				if(pstr[0]/*I[k*/ > sizeof name) {	//t@Cs
					stPTP.ResponseDataset.ResponseCode = PTP_ResponseCode_GeneralError;
					return 0;
				}
				get_string(name, pstr);
				strlwr(name);//P/ECE͏t@CgȂ̂ŁAIɏ邱ƂɂBK{ł͂ȂAȂƁA啶̃t@C]悤ƂɃG[ɂȂʓ|ł邩B
				if(is_protected_filename(name)) {//veNgt@C?
					stPTP.ResponseDataset.ResponseCode = PTP_ResponseCode_ObjectWriteProtected;
					return 0;
				}
				/* t@C쐬B */
				memcpy(stPTP.save_FRAM4, FRAM4_START, FRAM4_SIZE);	//FRAM4ޔ
				retval = pceFileCreate(name, size);			//FRAM4j󂳂
				memcpy(FRAM4_START, stPTP.save_FRAM4, FRAM4_SIZE);	//FRAM4𕜌
				if(retval) {	//t@C쐬s
					stPTP.ResponseDataset.ResponseCode = PTP_ResponseCode_StoreFull;
					return 0;
				}
			}
			{
				const pffsMASTERBLOCK* msb = (pffsMASTERBLOCK*)pceSystemGetInfo()->pffs_top;
				const DIRECTORY *pdir = msb->dir, *pdir_end = msb->dir + MAXDIR;
				do {
				//	/* gp̃fBNgGgȂ΁c(\usr\PIECE\sysdev\pcekn\file.c Search() Q) */	//t@C̔r
				//	if((pdir->name[0] == 0) || (pdir->name[0] == -1)) { continue; }	//				//fł̂ŕsv
					/* ܍쐬t@CɈvc */
					if(!strcmp(pdir->name, name)) {
						/* fBNgGgւ̃|C^IuWFNgnhƂāAX|X̃p[^Ɋi[B */
						stPTP.ObjectHandle = (int)pdir;	//SendObjectInfoSendObjectւ̈np
						stPTP.ResponseDataset.Parameter[0] = StorageID;			//Responder StorageID in which the object will be stored
					//	stPTP.ResponseDataset.Parameter[1] = -1;			//Responder Parent ObjectHandle in which the object will be stored
						stPTP.ResponseDataset.Parameter[2] = stPTP.ObjectHandle;	//Responderfs reserved ObjectHandle for the incoming object
						return 0;
					}
				} while(++pdir < pdir_end);
				/* ܍쐬t@C̃fBNgGgȂ(?)Bۂɂ͋N蓾Ȃ͂B */
				stPTP.ResponseDataset.ResponseCode = PTP_ResponseCode_GeneralError;
			}
		}
		return 0;
	/*--------------------------------------------------------------------------*/
	case PTP_OperationCode_SendObject:
		{
			int sct = stPTP.DataPos / BLOCKSIZE;			//݂̃f[^ʒuɑΉZN^ԍ
			int pos = stPTP.DataPos % BLOCKSIZE;			//[Nobt@݈̏ʒu
			do {
				int cnt = BLOCKSIZE - pos;			//[Nobt@ɏރoCg߂
				if(cnt > len) { cnt = len; }			//
				memcpy(&stPTP.WorkBuf[pos], buf, cnt);		//[Nobt@ɏ
				if(((          pos + cnt) ==     BLOCKSIZE) ||	//[Nobt@tɂȂ,,
				   ((stPTP.DataPos + len) == stPTP.DataLen)) {	//Ō̃pPbgMc
					const DIRECTORY* pdir = (DIRECTORY*)stPTP.ObjectHandle;	//SendObjectInfoSendObjectւ̈np
					FILEACC fa;
					if(pceFileOpen(&fa, pdir->name, FOMD_WR)) {	//IuWFNgnhs
						stPTP.ResponseDataset.ResponseCode = PTP_ResponseCode_InvalidObjectHandle;
						return 0;
					}
					/* [Nobt@ZN^֏ */
					memcpy(stPTP.save_FRAM4, FRAM4_START, FRAM4_SIZE);	//FRAM4ޔ
					pceFileWriteSct(&fa, stPTP.WorkBuf, sct, pos + cnt);	//FRAM4j󂳂
					memcpy(FRAM4_START, stPTP.save_FRAM4, FRAM4_SIZE);	//FRAM4𕜌
					pceFileClose(&fa);
					sct++;					//ZN^ԍ֐i߂
					pos = 0;				//[Nobt@݈̏ʒu擪֖߂
				}
				buf += cnt;					//[Nobt@ɏ񂾕AMf[^i߂
				len -= cnt;					//Mf[^̎coCg炷
			} while(len);						//Mf[^̎coCgLԁAJԂ
		}
		return 0;
	/*--------------------------------------------------------------------------*/
//	case PTP_OperationCode_SetDevicePropValue:
	/*--------------------------------------------------------------------------*/
	}
	/* ΁AG[ƂB */
	return -1;
}
/*--------------------------------------------------------------------------*/
static int process_datain(void* buf, int len) {
	/* f[^|C^ݒ肳ĂAf[^擾B */
	if(stPTP.DataPtr) { memcpy(buf, stPTP.DataPtr + stPTP.DataPos, len); }
	/* Iy[VR[hɂāc */
	switch(stPTP.OperationRequestDataset.OperationCode) {
	/*--------------------------------------------------------------------------*/
	case PTP_OperationCode_GetDeviceInfo:
	case PTP_OperationCode_GetStorageIDs:
	case PTP_OperationCode_GetStorageInfo:
	case PTP_OperationCode_GetObjectHandles:
	case PTP_OperationCode_GetObjectInfo:
		/** no job **/
		return 0;
	/*--------------------------------------------------------------------------*/
	case PTP_OperationCode_GetObject:
		{
			/* 1pPbg擾xɃt@CJƂB͈PłB
			 * ۂɎƂA̕@łǏox͏[āA薳ƔfB */
			const DIRECTORY* pdir = (DIRECTORY*)stPTP.OperationRequestDataset.Parameter[0];
			FILE* fp = fopen(pdir->name, "rb");	//ydvzfopen()͔r䂳ĂȂB荞݊Oœfopen()gƁAclipio.c̃t@CI/O\̃obt@鋰ꂪLBW[𗘗pAvP[V́A荞݊Ofopen(),fclose()gꍇAIɊ荞݋֎~ɂB
			if(!fp) {	//IuWFNgnhs
				stPTP.ResponseDataset.ResponseCode = PTP_ResponseCode_InvalidObjectHandle;
				return 0;
			}
			fseek(fp, stPTP.DataPos, SEEK_SET);	//ɂ̓G[`FbNׂAƉ肵ďȗB
			fread(buf, 1, len, fp);			//ɂ̓G[`FbNׂAƉ肵ďȗB
			fclose(fp);
		}
		return 0;
	/*--------------------------------------------------------------------------*/
//	case PTP_OperationCode_GetThumb:
//	case PTP_OperationCode_GetDevicePropDesc:
//	case PTP_OperationCode_GetDevicePropValue:
//	case PTP_OperationCode_GetPartialObject:
	/*--------------------------------------------------------------------------*/
	}
	/* ΁AG[ƂB */
	return -1;
}
